"""The graphical items are footprint and board items that are outside of the connectivity items.
   This includes graphical items on technical, user, and copper layers. Graphical items are also
   used to define complex pad geometries.

Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    10.02.2022 - created

Documentation taken from:
    https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_graphic_items
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, List

from kiutils.items.common import Effects, Position, RenderCache, Stroke
from kiutils.utils.strings import dequote

@dataclass
class GrText():
    """The ``gr_text`` token defines a graphical text.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_graphical_text
    """

    text: str = ""
    """The ``text`` attribute is a string that defines the text"""

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` defines the X and Y position coordinates and optional orientation angle of the text"""

    layer: Optional[str] = None
    """The ``layer`` token defines the canonical layer the text resides on"""

    effects: Effects = field(default_factory=lambda: Effects())
    """The ``effects`` token defines how the text is displayed"""

    tstamp: Optional[str] = None      # Used since KiCad 6
    """The ``tstamp`` token defines the unique identifier of the text object"""

    locked: bool = False
    """The ``locked`` token defines if the object may be moved or not"""

    renderCache: Optional[RenderCache] = None
    """If the ``effects`` token prescribe a TrueType font then the optional ``render_cache`` token 
    should be given in case the font can not be found on the current system.
    
    Available since KiCad v7"""

    @classmethod
    def from_sexpr(cls, exp: list) -> GrText:
        """Convert the given S-Expresstion into a GrText object

        Args:
            - exp (list): Part of parsed S-Expression ``(gr_text ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not gr_text

        Returns:
            - GrText: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'gr_text':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.text = exp[1]
        for item in exp[2:]:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                continue
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'effects': object.effects = Effects().from_sexpr(item)
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'render_cache': object.renderCache = RenderCache.from_sexpr(item)
        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        posA = f' {self.position.angle}' if self.position.angle is not None else ''
        layer =  f' (layer "{dequote(self.layer)}")' if self.layer is not None else ''
        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''
        locked = f' locked' if self.locked else ''

        expression =  f'{indents}(gr_text{locked} "{dequote(self.text)}" (at {self.position.X} {self.position.Y}{posA}){layer}{tstamp}\n'
        expression += f'{indents}  {self.effects.to_sexpr()}'
        if self.renderCache is not None:
            expression += self.renderCache.to_sexpr(indent+2)
        expression += f'{indents}){endline}'
        return expression

@dataclass
class GrTextBox():
    """The ``gr_text_box`` token defines a graphical rectangle containing line-wrapped text.

    Available since KiCad v7
    
    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_graphical_text_box
    """

    locked: bool = False
    """The ``locked`` token specifies if the text box can be moved. Defaults to ``False``."""

    text: str = ""
    """The ``text`` token defines the content of the text box. Defaults to an empty string."""

    start: Optional[Position] = None
    """The optional ``start`` token defines the top-left of a cardinally oriented text box"""

    end: Optional[Position] = None
    """The optional ``end`` token defines the bottom-right of a cardinally oriented text box"""

    pts: List[Position] = field(default_factory=list)
    """The ``pts`` token defines the four corners of a non-cardianlly oriented text box. The corners
    must be in order, but the winding can be either direction."""

    angle: Optional[float] = None
    """The optional ``angle`` token defines the rotation of the text box in degrees.

    Note:
        - If ``angle`` is not given, or is a cardinal angle (0, 90, 180 or 270), then the text box
          MUST have ``start`` and ``end`` tokens.
        - If ``angle`` is given and is not a cardinal angle, then the text box MUST have a ``pts``
          token (with 4 pts).
    """

    layer: str = "F.Cu"
    """The ``layer`` token defines the canonical layer the text box resides on. Defaults to
    ``F.Cu``."""

    tstamp: Optional[str] = None
    """The optional ``tstamp`` token defines the unique identifier of the text box"""

    effects: Optional[Effects] = None
    """The optional ``effects`` token describes the style of the text in the text box"""

    stroke: Optional[Stroke] = None
    """The optional ``stroke`` token describes the style of an optional border to be drawn around 
    the text box"""

    renderCache: Optional[RenderCache] = None
    """If the ``effects`` token prescribe a TrueType font then the optional ``render_cache`` token 
    should be given in case the font can not be found on the current system.
    
    Available since KiCad v7"""

    @classmethod
    def from_sexpr(cls, exp: list) -> GrTextBox:
        """Convert the given S-Expression into a GrTextBox object

        Args:
            - exp (list): Part of parsed S-Expression ``(gr_text_box ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not fp_text_box

        Returns:
            - GrTextBox: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list) or len(exp) < 2:
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'gr_text_box':
            raise Exception("Expression does not have the correct type")

        object = cls()

        # Extract "locked" token, if any is present
        if exp[1] == "locked" and not isinstance(exp[2], list):
            object.locked = True
            object.text = exp[2]
            start_at = 3
        else:
            object.text = exp[1]
            start_at = 2

        for item in exp[start_at:]:
            if item[0] == 'start': object.start = Position.from_sexpr(item)
            if item[0] == 'end': object.end = Position.from_sexpr(item)
            if item[0] == 'pts':
                for point in item[1:]:
                    object.pts.append(Position().from_sexpr(point))
            if item[0] == 'angle': object.angle = item[1]
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'effects': object.effects = Effects.from_sexpr(item)
            if item[0] == 'stroke': object.stroke = Stroke.from_sexpr(item)
            if item[0] == 'render_cache': object.renderCache = RenderCache.from_sexpr(item)

        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Raises:
            - Exception: When a non-cardinal angle is given and no corner points were defined using
              the ``self.pts`` token
            - Exception: When a cardinal angle or no angle is given and either start or end token
              is undefined

        Returns:
            - str: S-Expression of this object
        """
        if self.angle is not None and self.angle not in [0.0, 90.0, 180.0, 270.0]:
            if len(self.pts) != 4:
                raise Exception("None-cardinal angles must have exactly four corner points defined")
        if self.angle is None or self.angle in [0.0, 90.0, 180.0, 270.0]:
            if self.start is None or self.end is None:
                raise Exception("No angle or a cardinal angle needs a start and end token defined")

        indents = ' '*indent
        endline = '\n' if newline else ''

        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''
        angle = f'(angle {self.angle}) ' if self.angle is not None else ''
        start = f'(start {self.start.X} {self.start.Y}) ' if self.start is not None else ''
        end = f'(end {self.end.X} {self.end.Y}) ' if self.end is not None else ''
        locked = ' locked' if self.locked else ''

        expression = f'{indents}(gr_text_box{locked} "{dequote(self.text)}"\n'
        if len(self.pts) == 4:
            expression += f'{indents}  (pts\n'
            expression += f'{indents}    (xy {self.pts[0].X} {self.pts[0].Y})        (xy {self.pts[1].X} {self.pts[1].Y})        (xy {self.pts[2].X} {self.pts[2].Y})        (xy {self.pts[3].X} {self.pts[3].Y})\n'
            expression += f'{indents}  )\n'
        expression += f'{indents}  {start}{end}{angle}(layer "{dequote(self.layer)}"){tstamp}\n'
        if self.effects is not None:
            expression += self.effects.to_sexpr(indent+2)
        if self.stroke is not None:
            expression += self.stroke.to_sexpr(indent+2)
        if self.renderCache is not None:
            expression += self.renderCache.to_sexpr(indent+2)
        expression += f'{indents}){endline}'
        return expression

@dataclass
class GrLine():
    """The ``gr_line`` token defines a graphical line.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_graphical_line
    """

    start: Position = field(default_factory=lambda: Position())
    """The ``start`` token defines the coordinates of the start of the line"""

    end: Position = field(default_factory=lambda: Position())
    """The ``end`` token defines the coordinates of the end of the line"""

    angle: Optional[float] = None
    """The optional ``angle`` token defines the rotational angle of the line"""

    layer: Optional[str] = None
    """The ``layer`` token defines the canonical layer the rectangle resides on"""

    width: Optional[float] = 0.12     # Used for KiCad < 7
    """The ``width`` token defines the line width of the rectangle. (prior to version 7)"""

    tstamp: Optional[str] = None      # Used since KiCad 6
    """The ``tstamp`` token defines the unique identifier of the rectangle object"""

    locked: bool = False
    """The ``locked`` token defines if the object may be moved or not"""

    @classmethod
    def from_sexpr(cls, exp: list) -> GrLine:
        """Convert the given S-Expresstion into a GrLine object

        Args:
            - exp (list): Part of parsed S-Expression ``(gr_line ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not gr_line

        Returns:
            - GrLine: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'gr_line':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                continue
            if item[0] == 'start': object.start = Position.from_sexpr(item)
            if item[0] == 'end': object.end = Position.from_sexpr(item)
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'width': object.width = item[1]
        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        locked = f' locked' if self.locked else ''

        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''
        layer =  f' (layer "{dequote(self.layer)}")' if self.layer is not None else ''
        angle = f' (angle {self.angle}' if self.angle is not None else ''

        return f'{indents}(gr_line{locked} (start {self.start.X} {self.start.Y}) (end {self.end.X} {self.end.Y}){angle}{layer} (width {self.width}){tstamp}){endline}'

@dataclass
class GrRect():
    """The ``gr_rect`` token defines a graphical rectangle.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_graphical_rectangle
    """

    start: Position = field(default_factory=lambda: Position())
    """The ``start`` token defines the coordinates of the upper left corner of the rectangle"""

    end: Position = field(default_factory=lambda: Position())
    """The ``end`` token defines the coordinates of the low right corner of the rectangle"""

    layer: Optional[str] = None
    """The ``layer`` token defines the canonical layer the rectangle resides on"""

    width: Optional[float] = 0.12     # Used for KiCad < 7
    """The ``width`` token defines the line width of the rectangle. (prior to version 7)"""

    fill: Optional[str] = None
    """The optional ``fill`` toke defines how the rectangle is filled. Valid fill types are solid and none. If not defined, the rectangle is not filled"""

    tstamp: Optional[str] = None      # Used since KiCad 6
    """The ``tstamp`` token defines the unique identifier of the rectangle object"""

    locked: bool = False
    """The ``locked`` token defines if the object may be moved or not"""

    @classmethod
    def from_sexpr(cls, exp: list) -> GrRect:
        """Convert the given S-Expresstion into a GrRect object

        Args:
            - exp (list): Part of parsed S-Expression ``(gr_rect ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not gr_rect

        Returns:
            - GrRect: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'gr_rect':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                continue
            if item[0] == 'start': object.start = Position.from_sexpr(item)
            if item[0] == 'end': object.end = Position.from_sexpr(item)
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'fill': object.fill = item[1]
            if item[0] == 'width': object.width = item[1]
        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        locked = f' locked' if self.locked else ''

        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''
        layer =  f' (layer "{dequote(self.layer)}")' if self.layer is not None else ''
        fill = f' (fill {self.fill})' if self.fill is not None else ''

        return f'{indents}(gr_rect{locked} (start {self.start.X} {self.start.Y}) (end {self.end.X} {self.end.Y}){layer} (width {self.width}){fill}{tstamp}){endline}'

@dataclass
class GrCircle():
    """The ``gr_circle `` token defines a graphical circle.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_graphical_circle
    """

    center: Position = field(default_factory=lambda: Position())
    """The ``center`` token defines the coordinates of the center of the circle"""

    end: Position = field(default_factory=lambda: Position())
    """The ``end`` token defines the coordinates of the low right corner of the circle"""

    layer: Optional[str] = None
    """The ``layer`` token defines the canonical layer the circle resides on"""

    width: Optional[float] = 0.12     # Used for KiCad < 7
    """The ``width`` token defines the line width of the circle. (prior to version 7)"""

    fill: Optional[str] = None
    """The optional ``fill`` toke defines how the circle is filled. Valid fill types are solid and none. If not defined, the rectangle is not filled"""

    tstamp: Optional[str] = None      # Used since KiCad 6
    """The ``tstamp`` token defines the unique identifier of the circle object"""

    locked: bool = False
    """The ``locked`` token defines if the object may be moved or not"""

    @classmethod
    def from_sexpr(cls, exp: list) -> GrCircle:
        """Convert the given S-Expresstion into a GrCircle object

        Args:
            - exp (list): Part of parsed S-Expression ``(gr_circle ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not gr_circle

        Returns:
            - GrCircle: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'gr_circle':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                continue
            if item[0] == 'center': object.center = Position.from_sexpr(item)
            if item[0] == 'end': object.end = Position.from_sexpr(item)
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'fill': object.fill = item[1]
            if item[0] == 'width': object.width = item[1]

        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        locked = f' locked' if self.locked else ''

        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''
        layer =  f' (layer "{dequote(self.layer)}")' if self.layer is not None else ''
        fill = f' (fill {self.fill})' if self.fill is not None else ''

        return f'{indents}(gr_circle{locked} (center {self.center.X} {self.center.Y}) (end {self.end.X} {self.end.Y}){layer} (width {self.width}){fill}{tstamp}){endline}'

@dataclass
class GrArc():
    """The ``gr_arc`` token defines a graphic arc.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_graphical_arc
    """

    start: Position = field(default_factory=lambda: Position())
    """The ``start`` token defines the coordinates of the start position of the arc radius"""

    mid: Position = field(default_factory=lambda: Position())
    """The ``mid`` token defines the coordinates of the midpoint along the arc"""

    end: Position = field(default_factory=lambda: Position())
    """The ``end`` token defines the coordinates of the end position of the arc radius"""

    layer: Optional[str] = None
    """The ``layer`` token defines the canonical layer the arc resides on"""

    width: Optional[float] = 0.12     # Used for KiCad < 7
    """The ``width`` token defines the line width of the arc. (prior to version 7)"""

    tstamp: Optional[str] = None      # Used since KiCad 6
    """The ``tstamp`` token defines the unique identifier of the arc object."""

    locked: bool = False
    """The ``locked`` token defines if the object may be moved or not"""

    @classmethod
    def from_sexpr(cls, exp: list) -> GrArc:
        """Convert the given S-Expresstion into a GrArc object

        Args:
            - exp (list): Part of parsed S-Expression ``(gr_arc ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not gr_arc

        Returns:
            - GrArc: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'gr_arc':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                continue
            if item[0] == 'start': object.start = Position.from_sexpr(item)
            if item[0] == 'mid': object.mid = Position.from_sexpr(item)
            if item[0] == 'end': object.end = Position.from_sexpr(item)
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'width': object.width = item[1]

        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        locked = f' locked' if self.locked else ''

        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''
        layer =  f' (layer "{dequote(self.layer)}")' if self.layer is not None else ''

        return f'{indents}(gr_arc{locked} (start {self.start.X} {self.start.Y}) (mid {self.mid.X} {self.mid.Y}) (end {self.end.X} {self.end.Y}){layer} (width {self.width}){tstamp}){endline}'

@dataclass
class GrPoly():
    """The ``gr_poly`` token defines a graphic polygon in a footprint definition.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_graphical_polygon
    """

    layer: Optional[str] = None
    """The ``coordinates`` define the list of X/Y coordinates of the polygon outline"""

    coordinates: List[Position] = field(default_factory=list)
    """The ``layer`` token defines the canonical layer the polygon resides on"""

    width: Optional[float] = 0.12     # Used for KiCad < 7
    """The ``width`` token defines the line width of the polygon. (prior to version 7)"""

    fill: Optional[str] = None
    """The optional ``fill`` toke defines how the polygon is filled. Valid fill types are solid and none. If not defined, the rectangle is not filled"""

    tstamp: Optional[str] = None      # Used since KiCad 6
    """The ``tstamp`` token defines the unique identifier of the polygon object"""

    locked: bool = False
    """The ``locked`` token defines if the object may be moved or not"""

    @classmethod
    def from_sexpr(cls, exp: list) -> GrPoly:
        """Convert the given S-Expresstion into a GrPoly object

        Args:
            - exp (list): Part of parsed S-Expression ``(gr_poly ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not gr_poly

        Returns:
            - GrPoly: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'gr_poly':
            raise Exception("Expression does not have the correct type")

        object = cls()

        for item in exp:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                continue
            if item[0] == 'pts':
                for point in item[1:]:
                    object.coordinates.append(Position().from_sexpr(point))
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'fill': object.fill = item[1]
            if item[0] == 'width': object.width = item[1]

        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True, pts_newline: bool = False) -> str:
        """Generate the S-Expression representing this object. When no coordinates are set
        in the polygon, the resulting S-Expression will be left empty.

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.
            - pts_newline (bool): Adds a newline for the ``(pts ..)`` token as KiCad treats
                                  this different in Board files than Footprint files. Defaults to 
                                  False.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        if len(self.coordinates) == 0:
            return f'{indents}{endline}'

        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''
        layer =  f' (layer "{dequote(self.layer)}")' if self.layer is not None else ''
        fill = f' (fill {self.fill})' if self.fill is not None else ''
        locked = f' locked' if self.locked else ''

        if pts_newline:
            expression =  f'{indents}(gr_poly{locked}\n'
            expression += f'{indents}  (pts\n'
        else:
            expression =  f'{indents}(gr_poly{locked} (pts\n'

        for point in self.coordinates:
            expression += f'{indents}    (xy {point.X} {point.Y})\n'
        expression += f'{indents}  ){layer} (width {self.width}){fill}{tstamp}){endline}'
        return expression

@dataclass
class GrCurve():
    """The ``gr_curve`` token defines a graphic Cubic Bezier curve in a footprint definition.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_graphical_curve
    """
    coordinates: List[Position] = field(default_factory=list)
    """The ``coordinates`` define the list of X/Y coordinates of the curve outline"""

    layer: Optional[str] = None
    """The ``layer`` token defines the canonical layer the curve resides on"""

    width: Optional[float] = 0.12     # Used for KiCad < 7
    """The ``width`` token defines the line width of the curve. (prior to version 7)"""

    tstamp: Optional[str] = None      # Used since KiCad 6
    """The ``tstamp`` token defines the unique identifier of the curve object"""

    locked: bool = False
    """The ``locked`` token defines if the object may be moved or not"""

    @classmethod
    def from_sexpr(cls, exp: list) -> GrCurve:
        """Convert the given S-Expresstion into a GrCurve object

        Args:
            - exp (list): Part of parsed S-Expression ``(gr_curve ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not gr_curve

        Returns:
            - GrCurve: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'gr_curve':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                continue
            if item[0] == 'pts':
                for point in item[1:]:
                    object.coordinates.append(Position().from_sexpr(point))
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'width': object.width = item[1]

        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object. When no coordinates are set
        in the curve, the resulting S-Expression will be left empty.

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        if len(self.coordinates) == 0:
            return f'{indents}{endline}'

        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''
        layer =  f' (layer "{dequote(self.layer)}")' if self.layer is not None else ''
        locked = f' locked' if self.locked else ''

        expression = f'{indents}(gr_curve{locked} (pts\n'
        for point in self.coordinates:
            expression += f'{indents}  (xy {point.X} {point.Y})\n'
        expression += f'{indents}){layer} (width {self.width}{tstamp}){endline}'
        return expression



"""
Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    14.02.2022 - created

Documentation taken from:
    https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_symbols
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, List
from os import path
import re

from kiutils.items.common import Effects, Position, Property, Font
from kiutils.items.syitems import *
from kiutils.utils import sexpr
from kiutils.utils.strings import dequote
from kiutils.misc.config import KIUTILS_CREATE_NEW_VERSION_STR

@dataclass
class SymbolAlternativePin():
    pinName: str = ""
    """The ``pinName`` token defines the name of the alternative pin function"""

    electricalType: str = "input"
    """The ``electricalType`` defines the pin electrical connection. See symbol documentation for
    valid pin electrical connection types and descriptions."""

    graphicalStyle: str = "line"
    """The ``graphicalStyle`` defines the graphical style used to draw the pin. See symbol
    documentation for valid pin graphical styles and descriptions."""

    @classmethod
    def from_sexpr(cls, exp: list) -> SymbolAlternativePin:
        """Convert the given S-Expresstion into a SymbolAlternativePin object

        Args:
            - exp (list): Part of parsed S-Expression ``(alternate ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not alternate

        Returns:
            - SymbolAlternativePin: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'alternate':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.pinName = exp[1]
        object.electricalType = exp[2]
        object.graphicalStyle = exp[3]
        return object

    def to_sexpr(self, indent: int = 8, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 8.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        return f'{indents}(alternate "{dequote(self.pinName)}" {self.electricalType} {self.graphicalStyle}){endline}'

@dataclass
class SymbolPin():
    """The ``pin`` token defines a pin in a symbol definition.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_symbol_pin
    """

    electricalType: str = "input"
    """The ``electricalType`` defines the pin electrical connection. See documentation below for
    valid pin electrical connection types and descriptions."""

    graphicalStyle: str = "line"
    """The ``graphicalStyle`` defines the graphical style used to draw the pin. See documentation
    below for valid pin graphical styles and descriptions."""

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` defines the X and Y coordinates and rotation angle of the connection point
    of the pin relative to the symbol origin position"""

    length: float = 0.254
    """The ``length`` token attribute defines the LENGTH of the pin"""

    name: str = ""
    """The ``name`` token defines a string containing the name of the pin"""

    nameEffects: Optional[Effects] = None
    """The optional ``nameEffects`` token define how the pin's name is displayed. This token is
    mandatory for KiCad v6 and was made optional since KiCad v7."""

    number: str = "0"
    """The ``number`` token defines a string containing the NUMBER of the pin"""

    numberEffects: Optional[Effects] = None
    """The optional ``numberEffects`` token define how the pin's number is displayed. This token is
    mandatory for KiCad v6 and was made optional since KiCad v7."""

    hide: bool = False      # Missing in documentation
    """The 'hide' token defines if the pin should be hidden"""

    alternatePins: List[SymbolAlternativePin] = field(default_factory=list)
    """The 'alternate' token defines one or more alternative definitions for the symbol pin"""

    @classmethod
    def from_sexpr(cls, exp: list) -> SymbolPin:
        """Convert the given S-Expresstion into a SymbolPin object

        Args:
            - exp (list): Part of parsed S-Expression ``(pin ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not pin

        Returns:
            - SymbolPin: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'pin':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.electricalType = exp[1]
        object.graphicalStyle = exp[2]
        for item in exp[3:]:
            if type(item) != type([]):
                if item == 'hide': object.hide = True
                else: continue
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'length': object.length = item[1]
            if item[0] == 'name':
                object.name = item[1]
                if len(item) > 2:
                    object.nameEffects = Effects().from_sexpr(item[2])
            if item[0] == 'number':
                object.number = item[1]
                if len(item) > 2:
                    object.numberEffects = Effects().from_sexpr(item[2])
            if item[0] == 'alternate': object.alternatePins.append(SymbolAlternativePin().from_sexpr(item))
        return object

    def to_sexpr(self, indent: int = 4, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 4.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        newLineAdded = False

        hide = ' hide' if self.hide else ''
        posA = f' {self.position.angle}' if self.position.angle is not None else ''
        nameEffects = f' {self.nameEffects.to_sexpr(newline=False)}' if self.nameEffects is not None else ''
        numberEffects = f' {self.numberEffects.to_sexpr(newline=False)}' if self.numberEffects is not None else ''

        expression =  f'{indents}(pin {self.electricalType} {self.graphicalStyle} (at {self.position.X} {self.position.Y}{posA}) (length {self.length}){hide}'
        
        # Since KiCad v7 nightly: Missing name and number effects print both other tokens into 
        # the same line.
        # Constrained in: schematic/since_v7/test_symbolPinOptionalTokens
        if self.nameEffects is None and self.numberEffects is None:
            expression += f' (name "{dequote(self.name)}") (number "{dequote(self.number)}")'
        else:
            expression += f'\n{indents}  (name "{dequote(self.name)}"{nameEffects})\n'
            expression += f'{indents}  (number "{dequote(self.number)}"{numberEffects})\n'
            newLineAdded = True

        # Alternative pins always generate a line break
        if self.alternatePins:
            if not newLineAdded:
                expression += '\n' 
            newLineAdded = True
            for alternativePin in self.alternatePins:
                expression += alternativePin.to_sexpr(indent+2)

        if newLineAdded:
            expression += f'{indents}){endline}'
        else:
            expression += f'){endline}'
        return expression

@dataclass
class Symbol():
    """The ``symbol`` token defines a symbol or sub-unit of a parent symbol. There can be zero or more
    ``symbol`` tokens in a symbol library file.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_symbols
    """

    """Each symbol must have """
    @property
    def libId(self) -> str:
        """The ``lib_id`` token defines a unique "LIBRARY_ID" for each top level symbol in the
        library or a unique "UNIT_ID" for each unit embedded in a parent symbol. Library identifiers
        are only valid it top level symbols and unit identifiers are on valid as unit symbols inside
        a parent symbol. 
        
        The following conventions apply:
            - "LIBRARY_ID" (top-level symbol): ``[<libraryNickname>:]<entryName>`` (the library 
              nickname part is optional here)
            - "UNIT_ID" (child symbol): ``<entryName>_<unitId>_<styleId>``

        In ``kiutils``, the ``lib_id`` token is a combination of ``libraryNickname``, ``entryName``,
        ``unitId`` and ``styleId`` tokens. Setting the ``lib_id`` token will update all those tokens
        accordingly.

        Returns:
            - If the ``libraryNickname`` is set: ``<libraryNickname>:<entryName>``
            - If the ``libraryNickname`` is ``None``: ``<entryName>`` or ``<entryName>_<unitId>_<styleId>``,
              depending if these tokens are set.
        """
        if (self.unitId is not None and self.styleId is not None):
            unit_style_ids = f"_{self.unitId}_{self.styleId}"
        else:
            unit_style_ids = ""

        if self.libraryNickname:
            return f'{self.libraryNickname}:{self.entryName}'
        else:
            return f'{self.entryName}{unit_style_ids}'

    @libId.setter
    def libId(self, symbol_id: str):
        """Sets the ``lib_id`` token and parses its contents into the ``libraryNickname``,
        ``entryName``, ``unitId`` and ``styleId`` token. 

        See self.libId property description for more information.
        
        Args:
            - symbol_id (str): The symbol id in the following format: ``<libraryNickname>:<entryName>``,
              ``<entryName>_<unitId>_<styleId>`` or only ``<entryName>``, depending on if the symbol
              is a top-level symbol or a child symbol

        Raises:
            - Exception: If the given ID is neither a top-level nor a child symbol
        """
        # Try to parse the given ID
        parse_symbol_id = re.match(r"^(.+?):(.+?)$", symbol_id)
        if parse_symbol_id:
            # The symbol is a top-level symbol with a library nickname
            self.libraryNickname = parse_symbol_id.group(1)
            self.entryName = parse_symbol_id.group(2)
            self.unitId = None
            self.styleId = None
        else:
            parse_symbol_id = re.match(r"^(.+?)_(\d+?)_(\d+?)$", symbol_id)
            if parse_symbol_id:
                # The symbol is a child symbol
                self.libraryNickname = None
                self.entryName = parse_symbol_id.group(1)
                self.unitId = int(parse_symbol_id.group(2))
                self.styleId = int(parse_symbol_id.group(3))
            else:
                # The symbol is a top-level symbol without a library nickname
                self.libraryNickname = None
                self.entryName = symbol_id
                self.unitId = None
                self.styleId = None

        # Update units id to match parent id
        for unit in self.units:
            unit.entryName = self.entryName

    libraryNickname: Optional[str] = None
    """The optional ``libraryNickname`` token defines which symbol library this symbol belongs to
    and is a part of the ``id`` token"""
    
    entryName: str = None
    """The ``entryName`` token defines the actual name of the symbol and is a part of the ``id`` 
    token"""

    unitId: Optional[int] = None
    """The ``unitId`` token identifies which unit the symbol represents and is a part of 
    the ``id`` token"""

    styleId: Optional[int] = None
    """The ``styleId`` token indicates which body style the unit represents and is a part of the 
    ``id`` token"""

    extends: Optional[str] = None
    """The optional ``extends`` token attribute defines the "LIBRARY_ID" of another symbol inside the
    current library from which to derive a new symbol. Extended symbols currently can only have
    different symbol properties than their parent symbol."""

    hidePinNumbers: bool = False
    """The ``pin_numbers`` token defines the visibility setting of the symbol pin numbers for
    the entire symbol. If set to False, the all of the pin numbers in the symbol are visible."""

    pinNames: bool = False
    """The optional ``pinNames`` token defines the attributes for all of the pin names of the symbol.
    If the ``pinNames`` token is not defined, all symbol pins are shown with the default offset."""

    pinNamesHide: bool = False
    """The optional ``pinNamesOffset`` token defines the pin name of all pins should be hidden"""

    pinNamesOffset: Optional[float] = None
    """The optional ``pinNamesOffset`` token defines the pin name offset for all pin names of the
    symbol. If not defined, the pin name offset is 0.508mm (0.020")"""

    inBom: Optional[bool] = None
    """The optional ``inBom`` token, defines if a symbol is to be include in the bill of material
    output. If undefined, the token will not be generated in `self.to_sexpr()`."""

    onBoard: Optional[bool] = None
    """The ``onBoard`` token, defines if a symbol is to be exported from the schematic to the printed
    circuit board. If undefined, the token will not be generated in `self.to_sexpr()`."""

    # TODO: Describe this token
    isPower: bool = False           # Missing in documentation, added when "Als Spannungssymbol" is checked
    """The ``isPower`` token's documentation was not done yet .."""

    properties: List[Property] = field(default_factory=list)
    """The ``properties`` is a list of properties that define the symbol. The following properties are
    mandatory when defining a parent symbol: "Reference", "Value", "Footprint", and "Datasheet".
    All other properties are optional. Unit symbols cannot have any properties."""

    graphicItems: List = field(default_factory=list)
    """The ``graphicItems`` section is list of graphical arcs, circles, curves, lines, polygons, 
    rectangles, text and text boxes that define the symbol drawing. Possible items are defined in 
    ``kiutils.items.syitems``. This section can be empty if the symbol has no graphical items."""

    pins: List[SymbolPin] = field(default_factory=list)
    """The ``pins`` section is a list of pins that are used by the symbol. This section can be empty if
    the symbol does not have any pins."""

    units: List[Symbol] = field(default_factory=list)
    """The ``units`` can be one or more child symbol tokens embedded in a parent symbol"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Symbol:
        """Convert the given S-Expression into a Symbol object

        Args:
            - exp (list): Part of parsed S-Expression ``(symbol ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not symbol

        Returns:
            - Symbol: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'symbol':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.libId = exp[1]
        for item in exp[2:]:
            if item[0] == 'extends': object.extends = item[1]
            if item[0] == 'pin_numbers':
                if item[1] == 'hide':
                    object.hidePinNumbers = True
            if item[0] == 'pin_names':
                object.pinNames = True
                for property in item[1:]:
                    if type(property) == type([]):
                        if property[0] == 'offset': object.pinNamesOffset = property[1]
                    else:
                        if property == 'hide': object.pinNamesHide = True
            if item[0] == 'in_bom': object.inBom = True if item[1] == 'yes' else False
            if item[0] == 'on_board': object.onBoard = True if item[1] == 'yes' else False
            if item[0] == 'power': object.isPower = True

            if item[0] == 'symbol': object.units.append(Symbol().from_sexpr(item))
            if item[0] == 'property': object.properties.append(Property().from_sexpr(item))

            if item[0] == 'pin': object.pins.append(SymbolPin().from_sexpr(item))
            if item[0] == 'arc': object.graphicItems.append(SyArc().from_sexpr(item))
            if item[0] == 'circle': object.graphicItems.append(SyCircle().from_sexpr(item))
            if item[0] == 'curve': object.graphicItems.append(SyCurve().from_sexpr(item))
            if item[0] == 'polyline': object.graphicItems.append(SyPolyLine().from_sexpr(item))
            if item[0] == 'rectangle': object.graphicItems.append(SyRect().from_sexpr(item))
            if item[0] == 'text': object.graphicItems.append(SyText().from_sexpr(item))
            if item[0] == 'text_box': object.graphicItems.append(SyTextBox().from_sexpr(item))

        return object

    @classmethod
    def create_new(cls, id: str, reference: str, value: str,
                        footprint: str = "", datasheet: str = "") -> Symbol:
        """Creates a new empty symbol as KiCad would create it

        Args:
            - id (str): ID token of the symbol
            - reference (str): Reference designator
            - value (str): Value of the ``value`` property
            - footprint (str): Value of the ``footprint`` property. Defaults to "" (empty string).
            - datasheet (str): Value of the ``datasheet`` property. Defaults to "" (empty string).

        Returns:
            - Symbol: New symbol initialized with default values
        """
        symbol = cls()
        symbol.inBom = True
        symbol.onBoard = True
        symbol.libId = id
        symbol.properties.extend(
            [
                Property(key = "Reference", value = reference, id = 0,
                         effects = Effects(font=Font(width=1.27, height=1.27))),
                Property(key = "Value", value = value, id = 1,
                         effects = Effects(font=Font(width=1.27, height=1.27))),
                Property(key = "Footprint", value = footprint, id = 2,
                         effects = Effects(font=Font(width=1.27, height=1.27), hide=True)),
                Property(key = "Datasheet", value = datasheet, id = 3,
                         effects = Effects(font=Font(width=1.27, height=1.27), hide=True))
            ]
        )
        return symbol

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        obtext, ibtext = '', ''

        if self.inBom is not None:
            ibtext = 'yes' if self.inBom else 'no'
        inbom = f' (in_bom {ibtext})' if self.inBom is not None else ''
        if self.onBoard is not None:
            obtext = 'yes' if self.onBoard else 'no'
        onboard = f' (on_board {obtext})' if self.onBoard is not None else ''
        power = f' (power)' if self.isPower else ''
        pnhide = f' hide' if self.pinNamesHide else ''
        pnoffset = f' (offset {self.pinNamesOffset})' if self.pinNamesOffset is not None else ''
        pinnames = f' (pin_names{pnoffset}{pnhide})' if self.pinNames else ''
        pinnumbers = f' (pin_numbers hide)' if self.hidePinNumbers else ''
        extends = f' (extends "{dequote(self.extends)}")' if self.extends is not None else ''

        expression =  f'{indents}(symbol "{dequote(self.libId)}"{extends}{power}{pinnumbers}{pinnames}{inbom}{onboard}\n'
        for item in self.properties:
            expression += item.to_sexpr(indent+2)
        for item in self.graphicItems:
            expression += item.to_sexpr(indent+2)
        for item in self.pins:
            expression += item.to_sexpr(indent+2)
        for item in self.units:
            expression += item.to_sexpr(indent+2)
        expression += f'{indents}){endline}'
        return expression

@dataclass
class SymbolLib():
    """A symbol library defines the common format of ``.kicad_sym`` files. A symbol library may contain
    zero or more symbols.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-symbol-lib/
    """
    version: str = KIUTILS_CREATE_NEW_VERSION_STR
    """The ``version`` token attribute defines the symbol library version using the YYYYMMDD date format"""

    generator: Optional[str] = None
    """The ``generator`` token attribute defines the program used to write the file"""

    symbols: List[Symbol] = field(default_factory=list)
    """The ``symbols`` token defines a list of zero or more symbols that are part of the symbol library"""

    filePath: Optional[str] = None
    """The ``filePath`` token defines the path-like string to the library file. Automatically set when
    ``self.from_file()`` is used. Allows the use of ``self.to_file()`` without parameters."""

    @classmethod
    def from_file(cls, filepath: str, encoding: Optional[str] = None) -> SymbolLib:
        """Load a symbol library directly from a KiCad footprint file (`.kicad_sym`) and sets the
        ``self.filePath`` attribute to the given file path.

        Args:
            - filepath (str): Path or path-like object that points to the file
            - encoding (str, optional): Encoding of the input file. Defaults to None (platform
                                        dependent encoding).

        Raises:
            - Exception: If the given path is not a file

        Returns:
            - SymbolLib: Object of the SymbolLib class initialized with the given KiCad symbol library
        """
        if not path.isfile(filepath):
            raise Exception("Given path is not a file!")

        with open(filepath, 'r', encoding=encoding) as infile:
            item = cls.from_sexpr(sexpr.parse_sexp(infile.read()))
            item.filePath = filepath
            return item

    @classmethod
    def from_sexpr(cls, exp: list) -> SymbolLib:
        """Convert the given S-Expresstion into a SymbolLib object

        Args:
            - exp (list): Part of parsed S-Expression ``(kicad_symbol_lib ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not kicad_symbol_lib

        Returns:
            - SymbolLib: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'kicad_symbol_lib':
            raise Exception("Expression does not have the correct type")

        object = cls()

        for item in exp[1:]:
            if item[0] == 'version': object.version = item[1]
            if item[0] == 'generator': object.generator = item[1]
            if item[0] == 'symbol': object.symbols.append(Symbol().from_sexpr(item))
        return object

    def to_file(self, filepath = None, encoding: Optional[str] = None):
        """Save the object to a file in S-Expression format

        Args:
            - filepath (str, optional): Path-like string to the file. Defaults to None. If not set,
                                        the attribute ``self.filePath`` will be used instead.
            - encoding (str, optional): Encoding of the output file. Defaults to None (platform
                                        dependent encoding).

        Raises:
            - Exception: If no file path is given via the argument or via `self.filePath`
        """
        if filepath is None:
            if self.filePath is None:
                raise Exception("File path not set")
            filepath = self.filePath

        with open(filepath, 'w', encoding=encoding) as outfile:
            outfile.write(self.to_sexpr())

    def to_sexpr(self, indent: int = 0, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression =  f'{indents}(kicad_symbol_lib (version {self.version}) (generator {self.generator})\n'
        for item in self.symbols:
            expression += f'{indents}{item.to_sexpr(indent+2)}'
        expression += f'{indents}){endline}'
        return expression



"""Defines all syntax that is shared across the symbol library, footprint library,
   schematic, board and work sheet file formats.

Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    02.02.2022 - created

Documentation taken from:
    https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_common_syntax
"""

from __future__ import annotations
from abc import ABC, abstractmethod

from dataclasses import dataclass, field
from typing import Optional, List, Dict

from kiutils.utils.strings import dequote

@dataclass
class Position():
    """The ``position`` token defines the positional coordinates and rotation of an object.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_position_identifier
    """

    X: float = 0.0
    """The ``X`` attribute defines the horizontal position of the object"""

    Y: float = 0.0
    """The ``Y`` attribute defines the vertical position of the object"""

    angle: Optional[float] = None
    """The optional ``angle`` attribute defines the rotational angle of the object. Not all
    objects have rotational position definitions. Symbol text angles are stored in tenths
    of a degree. All other angles are stored in degrees."""

    # TODO: What is this? Documentation does not tell ..
    unlocked: bool = False
    """The ``unlocked`` token's description has to be defined yet .."""

    @classmethod
    def from_sexpr(cls, exp: list) -> Position:
        """Convert the given S-Expresstion into a Position object

        Args:
            - exp (list): Part of parsed S-Expression ``(xxx ...)``

        Raises:
            - Exception: When the given expression is not of type ``list`` or the list is less than
                         3 items long

        Returns:
            - Position: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list) or len(exp) < 3:
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.X = exp[1]
        object.Y = exp[2]
        if len(exp) >= 4:
            # More than four components means X, Y, and either angle or unlocked are present
            if exp[3] != 'unlocked':
                object.angle = exp[3]

        for item in exp:
            if item == 'unlocked': object.unlocked = True

        return object

    def to_sexpr(self) -> str:
        """This object does not have a direct S-Expression representation."""
        raise NotImplementedError("This object does not have a direct S-Expression representation")


@dataclass
class Coordinate():
    """The ``coordinate`` token defines a three-dimentional position"""

    X: float = 0.0
    """The ``X`` token defines the position of the object on the x-axis"""

    Y: float = 0.0
    """The ``Y`` token defines the position of the object on the y-axis"""

    Z: float = 0.0
    """The ``Z`` token defines the position of the object on the z-axis"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Coordinate:
        """Convert the given S-Expresstion into a Coordinate object

        Args:
            - exp (list): Part of parsed S-Expression ``(xyz ...)``

        Raises:
            - Exception: When given parameter's type is not a list or the list is not 4 items long
            - Exception: When the first item of the list is not xyz

        Returns:
            - Coordinate: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list) or len(exp) != 4:
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'xyz':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.X = exp[1]
        object.Y = exp[2]
        object.Z = exp[3]
        return object

    def to_sexpr(self, indent=0, newline=False) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to False.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        return f'{indents}(xyz {self.X} {self.Y} {self.Z}){endline}'


@dataclass
class ColorRGBA():
    """The ``color`` token defines a RGBA color"""

    R: int = 0
    """The ``R`` token defines the red channel of the color"""

    G: int = 0
    """The ``G`` token defines the green channel of the color"""

    B: int = 0
    """The ``B`` token defines the blue channel of the color"""

    A: int = 0
    """The ``A`` token defines the alpha channel of the color"""

    precision: Optional[int] = None
    """Wether the output of ``to_sexpr()`` should have a set number of precision after the decimal
    point of the ``self.A`` attribute"""

    @classmethod
    def from_sexpr(cls, exp: list) -> ColorRGBA:
        """Convert the given S-Expresstion into a ColorRGBA object

        Args:
            - exp (list): Part of parsed S-Expression ``(color ...)``

        Raises:
            - Exception: When given parameter's type is not a list or the list is not 5 items long
            - Exception: When the first item of the list is not color

        Returns:
            - ColorRGBA: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list) or len(exp) != 5:
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'color':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.R = exp[1]
        object.G = exp[2]
        object.B = exp[3]
        object.A = exp[4]
        return object

    def to_sexpr(self, indent=0, newline=False) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to False.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        if self.precision is not None:
            alpha = f'{self.A:.{self.precision}f}'
        else:
            alpha = f'{self.A}'

        return f'{indents}(color {self.R} {self.G} {self.B} {alpha}){endline}'

@dataclass
class Stroke():
    """The ``stroke`` token defines how the outlines of graphical objects are drawn.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/#_stroke_definition
    """

    width: float = 0.0
    """The ``width`` token attribute defines the line width of the graphic object"""

    type: Optional[str] = None
    """The optional ``type`` token attribute defines the line style of the graphic object. Valid
    stroke line styles are:
    - ``dash``, ``dash_dot``, ``dash_dot_dot`` (since KiCad v7), ``dot``, ``default``, ``solid``
    """

    color: Optional[ColorRGBA] = None
    """The ``color`` token attributes define the line red, green, blue, and alpha color settings.
    Defaults to ``None`` and was made optional since KiCad 7."""

    @classmethod
    def from_sexpr(cls, exp: list) -> Stroke:
        """Convert the given S-Expresstion into a Stroke object

        Args:
            - exp (list): Part of parsed S-Expression ``(stroke ...)``

        Raises:
            - Exception: When given parameter's type is not a list or the list is not 4 items long
            - Exception: When the first item of the list is not stroke

        Returns:
            - Stroke: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list) or len(exp) < 2:
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'stroke':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                continue
            if item[0] == 'width': object.width = item[1]
            if item[0] == 'type':  object.type = item[1]
            if item[0] == 'color': object.color = ColorRGBA.from_sexpr(item)
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        color = f' {self.color.to_sexpr()}' if self.color is not None else ''
        the_type = f' (type {self.type})' if self.type is not None else ''
        return f'{indents}(stroke (width {self.width}){the_type}{color}){endline}'



@dataclass
class Font():
    """The ``font`` token attributes define how text is shown.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/#_text_effects
    """
    face: Optional[str] = None
    """The optional 'face' token indicates the font family. It should be a TrueType font family
    name or "KiCad Font" for the KiCad stroke font. (Kicad version 7)"""

    height: float = 1.0
    """The 'height' token attributes define the font's height"""

    width: float = 1.0
    """The 'width' token attributes define the font's width"""

    thickness: Optional[float] = None
    """The 'thickness' token attribute defines the line thickness of the font"""

    bold: bool = False
    """The 'bold' token specifies if the font should be bold"""

    italic: bool = False
    """The 'italic' token specifies if the font should be italicized"""

    lineSpacing: Optional[float] = None
    """The 'line_spacing' token specifies the spacing between lines as a ratio of standard
    line-spacing. (Not yet supported)"""

    color: Optional[ColorRGBA] = None
    """The optional ``color`` token specifies the color of the text element

    Available since KiCad v7"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Font:
        """Convert the given S-Expresstion into a Font object

        Args:
            - exp (list): Part of parsed S-Expression ``(font ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not font

        Returns:
            - Font: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'font':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'bold': object.bold = True
                if item == 'italic': object.italic = True
                continue
            if item[0] == 'face': object.face = item[1]
            if item[0] == 'size':
                object.height = item[1]
                object.width = item[2]
            if item[0] == 'thickness': object.thickness = item[1]
            if item[0] == 'line_spacing': object.lineSpacing = item[1]
            if item[0] == 'color': object.color = ColorRGBA.from_sexpr(item)
        return object

    def to_sexpr(self, indent=0, newline=False) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to False.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        face_name, thickness, bold, italic, linespacing, color = '', '', '', '', '', ''

        if self.face is not None:        face_name = f'(face "{dequote(self.face)}") '
        if self.thickness is not None:   thickness = f' (thickness {self.thickness})'
        if self.bold == True:            bold = ' bold'
        if self.italic == True:          italic = ' italic'
        if self.lineSpacing is not None: linespacing = f' (line_spacing {self.lineSpacing})'
        if self.color is not None:       color = f' {self.color.to_sexpr()}'

        expression = f'{indents}(font {face_name}(size {self.height} {self.width}){color}{thickness}{bold}{italic}{linespacing}){endline}'
        return expression

@dataclass
class Justify():
    """The ``justify`` token defines the justification of a text object

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/#_text_effects
    """

    horizontally: Optional[str] = None
    """The ``horizontally`` token sets the horizontal justification. Valid values are ``right`` or ``left``"""

    vertically: Optional[str] = None
    """The ``vertically`` token sets the vertical justification. Valid values are ``top`` or ``bottom``"""

    mirror: bool = False
    """The ``mirror`` token defines if the text is mirrored or not"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Justify:
        """Convert the given S-Expresstion into a Justify object

        Args:
            - exp (list): Part of parsed S-Expression ``(justify ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not justify

        Returns:
            - Justify: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'justify':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            # 'center' is the standard on vertical but not on horizontal in work sheets
            if item == 'left' or item == 'right' or item == 'center': object.horizontally = item
            if item == 'top' or item == 'bottom': object.vertically = item
            if item == 'mirror': object.mirror = True
        return object

    def to_sexpr(self, indent=0, newline=False) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to False.

        Returns:
            - str: S-Expression of this object or an empty string (depending on given indentation
              and newline settings) if no justification is given. This will cause the text to be
              horizontally and vertically aligend
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        if self.horizontally is None and self.vertically is None and self.mirror == False:
            return f'{indents}{endline}';

        horizontally, vertically, mirror = '', '', ''

        if self.horizontally is not None: horizontally = f' {self.horizontally}'
        if self.vertically is not None: vertically = f' {self.vertically}'
        if self.mirror: mirror = f' mirror'

        expression = f'{indents}(justify{horizontally}{vertically}{mirror}){endline}'
        return expression

@dataclass
class Effects():
    """All text objects can have an optional effects section that defines how the text is displayed.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/#_text_effects
    """

    font: Font = field(default_factory=lambda: Font())
    """The ``font`` token defines how the text is shown"""

    justify: Justify = field(default_factory=lambda: Justify())
    """The ``justify`` token defines the justification of the text"""

    hide: bool = False
    """The optional ``hide`` token defines if the text is hidden"""

    href: Optional[str] = None
    """The optional ``href`` token specifies a link that the text element represents.

    Available since KiCad v7"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Effects:
        """Convert the given S-Expresstion into a Effects object

        Args:
            - exp (list): Part of parsed S-Expression ``(effects ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not effects

        Returns:
            - Effects: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'effects':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if type(item) != type([]):
                if item == 'hide': object.hide = True
                else: continue
            if item[0] == 'font': object.font = Font().from_sexpr(item)
            if item[0] == 'justify': object.justify = Justify().from_sexpr(item)
            if item[0] == 'href': object.href = item[1]
        return object

    def to_sexpr(self, indent=0, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        justify = f' {self.justify.to_sexpr()}' if self.justify.to_sexpr() != '' else ''
        hide = f' hide' if self.hide else ''
        href = f' (href "{dequote(self.href)}")' if self.href is not None else ''

        expression =  f'{indents}(effects {self.font.to_sexpr()}{justify}{href}{hide}){endline}'
        return expression


@dataclass
class Net():
    """The ``net`` token defines the number and name of a net"""

    number: int = 0
    """The ``number`` token defines the integer number of the net"""

    name: str = ""
    """The ``name`` token defines the name of the net"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Net:
        """Convert the given S-Expresstion into a Net object

        Args:
            - exp (list): Part of parsed S-Expression ``(net ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not net

        Returns:
            - Net: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'net':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.number = exp[1]
        object.name = exp[2]
        return object

    def to_sexpr(self, indent: int = 0, newline: bool = False) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to False.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        return f'{indents}(net {self.number} "{dequote(self.name)}"){endline}'

@dataclass
class Group():
    """The ``group`` token defines a group of items.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_group
    """

    name: str = ""
    """The ``name`` attribute defines the name of the group"""

    locked: bool = False
    """The ``locked`` token defines if the group may be moved or not"""

    id: str = ""
    """The ``id`` token attribute defines the unique identifier of the group"""

    members: List[str] = field(default_factory=list)
    """The ``members`` token attributes define a list of unique identifiers of the objects belonging to the group"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Group:
        """Convert the given S-Expresstion into a Group object

        Args:
            - exp (list): Part of parsed S-Expression ``(group ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not group

        Returns:
            - Group: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'group':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.name = exp[1]
        for item in exp:
            if type(item) != type([]):
                if item == 'locked': object.locked = True
                continue
            if item[0] == 'id': object.id = item[1]
            if item[0] == 'members':
                for member in item[1:]:
                    object.members.append(member)
        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        locked = f' locked' if self.locked else ''

        expression =  f'{indents}(group "{dequote(self.name)}"{locked} (id {self.id})\n'
        expression += f'{indents}  (members\n'
        for member in self.members:
            expression += f'{indents}    {member}\n'

        expression += f'{indents}  )\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class PageSettings():
    """The ``paper`` token defines the drawing page size and orientation.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/#_page_settings
    """
    paperSize: str = "A4"
    """The ``paperSize`` token defines the size of the paper. Valid sizes are `A0`, `A1`, `A2`,
    `A3`, `A4`, `A5`, ``A``, ``B``, ``C``, ``D`` and ``E``. When using user-defines page sizes, set
    this to ``User``"""

    width: Optional[float] = None
    """The ``width`` token sets the width of a user-defines page size"""

    height: Optional[float] = None
    """The ``height`` token sets the height of a user-defines page size"""

    portrait: bool = False
    """The ``portrait`` token defines if the page is in portrait or landscape mode"""

    @classmethod
    def from_sexpr(cls, exp: list) -> PageSettings:
        """Convert the given S-Expresstion into a PageSettings object

        Args:
            - exp (list): Part of parsed S-Expression ``(paper ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not paper
            - Exception: When the paper type is set to ``User`` and the list's length is not 4

        Returns:
            - PageSettings: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'paper':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.paperSize = exp[1]
        if object.paperSize == "User":
            if len(exp) < 4:
                raise Exception("PageSettings: Expected more data for paper type 'User'")

            object.width = exp[2]
            object.height = exp[3]
        for item in exp:
            if type(item) != type([]):
                if item == 'portrait': object.portrait = True
                continue
        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Raises:
            - Exception: When paper size is set to ``User`` and width or height is not specified

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        width, height = '', ''
        portrait = ' portrait' if self.portrait else ''
        if self.paperSize == 'User':
            if self.width is None or self.height is None:
                raise Exception("Page size set to 'User' but width or height not specified")
            width = f' {self.width}'
            height = f' {self.height}'
        return f'{indents}(paper "{dequote(self.paperSize)}"{width}{height}{portrait}){endline}'

@dataclass
class TitleBlock():
    """The ``title_block`` token defines the contents of the title block.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/#_title_block
    """

    title: Optional[str] = None
    """The optional ``title`` token attribute is a quoted string that defines the document title"""

    date: Optional[str] = None
    """The optional ``date`` token attribute is a quoted string that defines the document date using the YYYY-MM-DD format"""

    revision: Optional[str] = None
    """The optional ``revision`` token attribute is a quoted string that defines the document revision"""

    company: Optional[str] = None
    """The optional ``company`` token attribute is a quoted string that defines the document company name"""

    comments: Dict[int, str] = field(default_factory=dict)
    """The ``comments`` token attributes define a dictionary of document comments where the key is
    a number from 1 to 9 and the value is a comment string"""

    @classmethod
    def from_sexpr(cls, exp: list) -> TitleBlock:
        """Convert the given S-Expresstion into a TitleBlock object

        Args:
            - exp (list): Part of parsed S-Expression ``(title_block ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not title_block

        Returns:
            - TitleBlock: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'title_block':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'title': object.title = item[1]
            if item[0] == 'date': object.date = item[1]
            if item[0] == 'rev': object.revision = item[1]
            if item[0] == 'company': object.company = item[1]
            if item[0] == 'comment': object.comments.update({item[1]: item[2]})
        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression =  f'{indents}(title_block\n'
        if self.title is not None:
            expression += f'{indents}  (title "{dequote(self.title)}")\n'

        if self.date is not None:
            expression += f'{indents}  (date "{dequote(self.date)}")\n'

        if self.revision is not None:
            expression += f'{indents}  (rev "{dequote(self.revision)}")\n'

        if self.company is not None:
            expression += f'{indents}  (company "{dequote(self.company)}")\n'

        for number, comment in self.comments.items():
            expression += f'{indents}  (comment {number} "{dequote(comment)}")\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class Property():
    """The ``property`` token defines a symbol property when used inside a ``symbol`` definition.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_symbol_property
    """

    key: str = ""
    """The ``key`` string defines the name of the property and must be unique"""

    value: str = ""
    """The ``value`` string defines the value of the property"""

    id: Optional[int] = None
    """The ``id`` token defines an integer ID for the property and must be unique.

    Optional since KiCad v7, but required in older versions"""

    position: Position = field(default_factory=lambda: Position(angle=0))
    """The ``position`` defines the X and Y coordinates as well as the rotation angle of the property.
    All three items will initially be set to zero."""

    effects: Optional[Effects] = None
    """The optional ``effects`` section defines how the text is displayed"""

    showName: bool = False
    """The ``show_name`` token defines if the property name is visibly shown. Used for netclass
    labels.

    Available since KiCad v7"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Property:
        """Convert the given S-Expresstion into a Property object

        Args:
            - exp (list): Part of parsed S-Expression ``(property ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not property

        Returns:
            - Property: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'property':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.key = exp[1]
        object.value = exp[2]
        for item in exp[3:]:
            if item[0] == 'id': object.id = item[1]
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'effects': object.effects = Effects().from_sexpr(item)
            if item[0] == 'show_name': object.showName = True
        return object

    def to_sexpr(self, indent: int = 4, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 4.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        posA = f' {self.position.angle}' if self.position.angle is not None else ''
        id = f' (id {self.id})' if self.id is not None else ''
        sn = ' (show_name)' if self.showName else ''

        expression =  f'{indents}(property "{dequote(self.key)}" "{dequote(self.value)}"{id} (at {self.position.X} {self.position.Y}{posA}){sn}'
        if self.effects is not None:
            expression += f'\n{self.effects.to_sexpr(indent+2)}'
            expression += f'{indents}){endline}'
        else:
            expression += f'){endline}'
        return expression

@dataclass
class RenderCachePolygon():
    """A polygon used by the ``render_cache`` token

    Used since KiCad v7
    """

    pts: List[Position] = field(default_factory=list)
    """The ``pts`` token defines a list of points that define the outlines of the polygon"""

    @classmethod
    def from_sexpr(cls, exp: list) -> RenderCachePolygon:
        """Convert the given S-Expresstion into a RenderCachePolygon object

        Args:
            - exp (list): Part of parsed S-Expression ``(polygon ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not polygon

        Returns:
            - RenderCachePolygon: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'polygon':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'pts':
                for point in item[1:]:
                    object.pts.append(Position.from_sexpr(point))
        return object

    def to_sexpr(self, indent: int = 6, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 6.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression = f'{indents}(polygon\n'
        expression += f'{indents}  (pts'

        for i, point in enumerate(self.pts):
            if i % 4 == 0:
                expression += f'\n'
            expression += f'{indents}    '
            expression += f'(xy {point.X} {point.Y})'

        # NOTE: This expects the length of the points array to be a multiple of four to get the
        #       formatting right.
        expression += f'\n{indents}  )\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class RenderCache():
    """The ``render_cache`` token defines a cache for none-standard fonts.

    Used since KiCad v7

    Documentation:
        - None found (05.03.2023), seems to be used in ''text_box'' tokens for custom fonts
    """

    text: str = ""
    """The ``text`` token defines which text the cache represents. Defaults to an empty string."""

    id: int = 0
    """The ``id`` token is some number after the text. Defaults to 0."""

    polygons: List[Position] = field(default_factory=list)
    """The ``polygons`` token is a list of polygons that define the outline of the cached text"""

    @classmethod
    def from_sexpr(cls, exp: list) -> RenderCache:
        """Convert the given S-Expresstion into a RenderCache object

        Args:
            - exp (list): Part of parsed S-Expression ``(render_cache ...)``

        Raises:
            - Exception: When given parameter's type is not a list or the list is smaller than 3
            - Exception: When the first item of the list is not render_cache

        Returns:
            - RenderCache: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list) or len(exp) < 3:
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'render_cache':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.text = exp[1]
        object.id = exp[2]
        for item in exp:
            if item[0] == 'polygon': object.polygons.append(RenderCachePolygon.from_sexpr(item))
        return object

    def to_sexpr(self, indent: int = 4, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 4.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression = f'{indents}(render_cache "{dequote(self.text)}" {self.id}\n'
        for poly in self.polygons:
            expression += poly.to_sexpr(indent+2)
        expression += f'{indents}){endline}'
        return expression

@dataclass
class Fill():
    """The ``fill`` token defines how schematic and symbol graphical items are filled

    Documentation:
        - https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_fill_definition
    """

    type: str = "none"
    """The ``type`` attribute defines how the graphical item is filled. Defaults to ``None``.
    Possible values are:
    - ``none``: Graphic is not filled
    - ``outline``: Graphic item filled with the line color
    - ``background``: Graphic item filled with the theme background color"""

    color: Optional[ColorRGBA] = None
    """The optional ``color`` token defines the color of the filled item.

    Available since KiCad v7"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Fill:
        """Convert the given S-Expresstion into a Fill object

        Args:
            - exp (list): Part of parsed S-Expression ``(fill ...)``

        Raises:
            - Exception: When given parameter's type is not a list or the list is smaller than 3
            - Exception: When the first item of the list is not fill

        Returns:
            - Fill: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'fill':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'type': object.type = item[1]
            if item[0] == 'color': object.color = ColorRGBA().from_sexpr(item)
        return object

    def to_sexpr(self, indent: int = 4, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 4.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        color = f' {self.color.to_sexpr()}' if self.color is not None else ''

        expression = f'{indents}(fill (type {self.type}){color}){endline}'
        return expression

@dataclass
class Image():
    """The ``image`` token defines an image embedded into the file

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_image_section
    """

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` defines the X and Y coordinates of the image"""

    scale: Optional[float] = None
    """The optional ``scale`` token attribute defines the scale factor (size) of the image"""

    data: List[str] = field(default_factory=list)
    """The ``data`` token attribute defines the image data in the portable network graphics
    format (PNG) encoded with MIME type base64 as a list of strings"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""

    layer: Optional[str] = None
    """The optional ``layer`` token defines the canonical layer name when the image is used inside
    a footprint or PCB. When used inside a schematic, this token is required to be ``None``."""

    @classmethod
    def from_sexpr(cls, exp: list) -> Image:
        """Convert the given S-Expresstion into a Image object

        Args:
            - exp (list): Part of parsed S-Expression ``(image ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not image

        Returns:
            - Image: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'image':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'scale': object.scale = item[1]
            if item[0] == 'uuid': object.uuid = item[1]
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'data':
                for b64part in item[1:]:
                    object.data.append(b64part)
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        scale = f' (scale {self.scale})' if self.scale is not None else ''
        layer = f' (layer "{dequote(self.layer)}")' if self.layer is not None else ''

        expression =  f'{indents}(image (at {self.position.X} {self.position.Y}){layer}{scale}\n'
        if self.uuid is not None:
            expression += f'{indents}  (uuid {self.uuid})\n'
        expression += f'{indents}  (data\n'
        for b64part in self.data:
            expression += f'{indents}    {b64part}\n'
        expression += f'{indents}  )\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class ProjectInstance(ABC):
    """The ``instances`` token defines a project instance and serves as an abstract base class for
    symbol and hierarchical sheet project instances.
    
    Available since KiCad v7."""

    name: str = ""
    """The ``name`` token defines the name of the project instance"""

    @abstractmethod
    def from_sexpr(cls, exp: list) -> ProjectInstance:
        raise NotImplementedError

    @abstractmethod
    def to_sexpr(self, indent=2, newline=True) -> str:
        raise NotImplementedError
